{****************************************************}
{* Common routines unit for LogicGates4 application *}
{****************************************************}

unit common;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, ExtCtrls, Graphics;

type
  TBinary = array[0..3] of Boolean;                                            // 4-bit binary number
  TSwitchData = record                                                         // switches data (cf. in subtractor units for details)
    Name: string;
    Data: array[0..4] of Integer;
  end;
  TSwitches = array of record                                                  // switches objects (transparent image and bar shapes)
    Switch0: TImage;
    Switch1, Switch2: TShape;
  end;
  TLeds = array of TShape;                                                     // LED objects (shape)
  TStatus = array of Boolean;                                                  // switch/LED status (on/off)

const
  Bits: array[False..True] of Integer = (0, 1);                                // used to convert Boolean to binary 0/1
  SUB_Digits: array[0..9] of string = (                                        // UTF-8 number subscripts
    #$E2#$82#$80, #$E2#$82#$81, #$E2#$82#$82, #$E2#$82#$83, #$E2#$82#$84,
    #$E2#$82#$85, #$E2#$82#$86, #$E2#$82#$87, #$E2#$82#$88, #$E2#$82#$89
  );

// ----------------------------------------------- Public procedures ------------------------------------------------

function Complement2(Bin: TBinary): TBinary;
procedure SwitchOnOff(DSwitches: array of TSwitchData; Switch0: TImage; Switch1, Switch2: TShape; SwitchOn: Boolean);
procedure SwitchesOnOff(DSwitches: array of TSwitchData; Switches: TSwitches; Status: TStatus);
procedure LedOnOff(Led: TShape; LedOn: Boolean);
procedure LedsOnOff(Leds: TLeds; Status: TStatus);
procedure HalfSubtractor(A, B: Boolean; out S, C: Boolean);
procedure Subtractor(A, B, C: Boolean; out C1, C2, C3, Sout, Cout: Boolean);
procedure Adder(A, B, C: Boolean; out S1, C1, S2, C2, Sout, Cout: Boolean);

// ------------------------------------------------------------------------------------------------------------------

implementation

{ Calculate 2'complement of 4-bit binary number }

function Complement2(Bin: TBinary): TBinary;

var
  Sum, Carry, I: Integer;
  Compl: TBinary;

begin
  // Calculate 1's complement
  for I := 0 to 3 do
    Bin[I] := not Bin[I];
  // Add binary 1 to form 2's complement
  Carry := 1;
  for I := 3 downto 0 do begin
    Compl[I] := False;
    Sum := Bits[Bin[I]] + Carry;
    if Sum <= 1 then
      Carry := 0
    else begin
      Sum := 0;
      Carry := 1;
    end;
    if Sum = 1 then
      Compl[I] :=  True;
  end;
  Result := Compl;
end;

{ Set given switch to position "on" or "off" }

procedure SwitchOnOff(DSwitches: array of TSwitchData; Switch0: TImage; Switch1, Switch2: TShape; SwitchOn: Boolean);

var
  I, IX: Integer;

begin
  // Determine index for switch data array by looking up the actual switch image object's name
  for I := 0 to Length(DSwitches) - 1 do begin
    if DSwitches[I].Name = Copy(Switch0.Name, 3, Length(Switch0.Name)) then
      IX := I;
  end;
  // Switch that moves horizontally
  if DSwitches[IX].Data[3] <> 0 then begin
    // Switch's horizontal "off" position (defined in its data)
    Switch0.Left := DSwitches[IX].Data[0]; Switch1.Left := DSwitches[IX].Data[1]; Switch2.Left := DSwitches[IX].Data[2];
    if SwitchOn then begin
      // Adjust switch object's horizontal position if the switch has to be set "on"
      Switch0.Left := Switch0.Left + DSwitches[IX].Data[3] * 10;
      Switch1.Left := Switch1.Left + DSwitches[IX].Data[3] * 10;
      Switch2.Left := Switch2.Left + DSwitches[IX].Data[3] * 10;
    end;
  end
  // Switch that moves vertically
  else if DSwitches[IX].Data[4] <> 0 then begin
    // Switch's vertical "off" position (defined in its data)
    Switch0.Top := DSwitches[IX].Data[0]; Switch1.Top := DSwitches[IX].Data[1]; Switch2.Top := DSwitches[IX].Data[2];
    if SwitchOn then begin
      // Adjust switch object's vertical position if the switch has to be set "on"
      Switch0.Top := Switch0.Top + DSwitches[IX].Data[4] * 10;
      Switch1.Top := Switch1.Top + DSwitches[IX].Data[4] * 10;
      Switch2.Top := Switch2.Top + DSwitches[IX].Data[4] * 10;
    end;
  end;
end;

{ Set all switches to "on" or "off" position (depending on status values) }

procedure SwitchesOnOff(DSwitches: array of TSwitchData; Switches: TSwitches; Status: TStatus);

var
  I: Integer;

begin
  for I := 0 to Length(Switches) - 1 do
    // Set position for each switch (array element)
    SwitchOnOff(DSwitches, Switches[I].Switch0, Switches[I].Switch1, Switches[I].Switch2, Status[I]);
end;

{ Turn a given LED "on" or "off" }

procedure LedOnOff(Led: TShape; LedOn: Boolean);

begin
  if LedOn then
    // LED on shown by solid color
    Led.Brush.Style := bsSolid
  else
    // LED off shown by pattern
    Led.Brush.Style := bsDiagCross;
end;

{ Turn all LEDs "on" or "off" (depending on status values) }

procedure LedsOnOff(Leds: TLeds; Status: TStatus);

var
  I: Integer;

begin
  for I := 0 to Length(Leds) - 1 do begin
    // Turn each LED (array element) "on" or "off"
    LedOnOff(Leds[I], Status[I]);
  end;
end;

{ 1-bit half-adder }

procedure HalfAdder(A, B: Boolean; out S, C: Boolean);

// Calculation done as does the real circuit: Usage of an XOR and an AND gate

begin
  S := A xor B;
  C := A and B;
end;

{ 1-bit full-adder }

procedure Adder(A, B, C: Boolean; out S1, C1, S2, C2, Sout, Cout: Boolean);

// Calculation done as does the real circuit: Usage of two half-adders and an Or gate

begin
  HalfAdder(A, B, S1, C1);
  HalfAdder(C, S1, S2, C2);
  Sout := S2; Cout := C1 or C2;
end;

{ 1-bit-half-subtractor }

procedure HalfSubtractor(A, B: Boolean; out S, C: Boolean);

// Calculation done as does the real circuit: Usage of the XOR and AND gate (plus inverter)

begin
  S := A xor B;
  C := (not A) and B;                                                          // inverter here!
end;

{ 1-bit-full-subtractor }

procedure Subtractor(A, B, C: Boolean; out C1, C2, C3, Sout, Cout: Boolean);

// Calculation done as does the real circuit: Usage of the XOR, AND and OR gates (plus inverter)

begin
  C1 := (not A) and B; C2 := (not A) and C; C3 := B and C;                     // input A inverted!
  Sout := A xor B xor C; Cout := C1 or C2 or C3;
end;

end.

